
from .element import Element

class Button(Element):
    """This class is a UI element that works as a button. It can be selected, and
    an action can be acttached to it that is triggered when the button is activated.
    :param label: The text displayed inside the button
    :type label: str
    :param action: The action that the button activates
    :type action: function 
    :param refreshFunction: Function to refresh the label text on update
    :type refreshFunction: function, optional
    :param color: The text color to draw the element with
    :type color: int, optional
    :param setLabelToResult: Whether or not to set the label of the button to the 
    :return value of its action function
    :type setLabelToResult: bool, optional
    """

    def __init__(self, label, action, refreshFunction=None, color=0, setLabelToResult=False):
        """Constructor method
        """
        super().__init__(label, refreshFunction=refreshFunction, color=color)
        self.action = action
        self.selectable = True
        self.setLabelToResult = setLabelToResult

    def triggerAction(self):
        """Run the button's action function"""
        result = self.action()
        if type(result) == str and self.setLabelToResult:
            self.label = str(result)

    def getStr(self, selected=False):
        """Get the string that the button should display as
        :param selected: Whether or not the button is selected in the menu
        :type selected: bool, optional
        """
        if selected:
            return '[>' + self.label + '<]'
        else:
            return '[ ' + self.label + ' ]'



import curses
from . import colors

class Menu:
    """A class for storing different 'screens' of elements, as well as handling input on 
    these screens. The main purpose of this class is to easily switch between menus using
    the UIManager class
    :param name: The name of the menu
    :type name: str
    """
    def __init__(self, name):
        """Constructor method
        """
        self.name = name
        self.elements = {} 
        self.selectedIndex = 0
        self.hasSelectable = False

    def addElement(self, name, element):
        """Add an element to the menu
        :param name: The name of the element
        :type name: str
        :param element: The element to add
        :type element: ui.Element
        """
        if name in self.elements.keys():
            raise Exception(f"Element with name '{name}' already exists")
        self.elements[name] = element
        if element.selectable and self.hasSelectable == False:
            self.hasSelectable = True
            self.selectedIndex = len(self.elements)-1

    def handleInput(self, key):
        """Forward the input key to the selected element in the menu
        :param key: The input key
        :type key: int
        """
        elementKey = list(self.elements)[self.selectedIndex]
        if key == curses.KEY_UP or key == ord('k'): 
            if self.selectedIndex <= 0:
                return
            self.decreaseSelectedIndex()
        elif key == curses.KEY_DOWN or key == ord('j'):
            if self.selectedIndex >= len(self.elements)-1:
                return
            self.increaseSelectedIndex()
        elif key == 10: # Enter/Return
            self.elements[elementKey].triggerAction()
        elif key == ord('q'):
            exit()
        else:
            self.elements[elementKey].handleInput(key)

    def increaseSelectedIndex(self):
        """Increase the selected index of the menu, skipping unselectable elements"""
        if self.selectedIndex >= len(self.elements)-1:
            return
        self.selectedIndex += 1
        while self.elements[list(self.elements)[self.selectedIndex]].selectable == False:
            self.increaseSelectedIndex()
            if(self.selectedIndex >= len(self.elements)-1):
                self.decreaseSelectedIndex()
                break

    def decreaseSelectedIndex(self):
        """Decrease the selected index of the menu, skipping unselectable elements"""
        if self.selectedIndex <= 0:
            return
        self.selectedIndex -= 1
        while self.elements[list(self.elements)[self.selectedIndex]].selectable == False:
            self.decreaseSelectedIndex()
            if(self.selectedIndex <= 0):
                self.increaseSelectedIndex()
                break
            
            
    def update(self):
        """Run the update function of each element in the menu"""
        for key in self.elements.keys():
            self.elements[key].update()
    
    def display(self, stdscr, ending='\n'):
        """Display all of the elements in the menu
        :param stdscr: The display to print to
        :param ending: The character(s) to print after each element
        :type ending: str
        """
        stdscr.clear()
        for id, key in enumerate(self.elements.keys()):
            element = self.elements[key]
            selected = (id == self.selectedIndex)
            
            if element.isContainer:
                separator = element.separator
                for barid, e in enumerate(element.elements):
                    barselected = False
                    if barid == element.selectedIndex:
                        barselected = True
                    color = e.color
                    augments = curses.color_pair(color)

                    if selected:
                        augments = curses.color_pair(color + colors.BG_GRAY)

                    stdscr.addstr(str(e.getStr(selected=(barselected & selected))) + separator, augments)
                stdscr.addstr(' \n') 
            else:
                color = element.color
                augments = curses.color_pair(color)
                if selected:
                    augments = curses.color_pair(color + colors.BG_GRAY)
                    augments |= curses.A_UNDERLINE
                stdscr.addstr(str(element.getStr(selected=selected)) + ending, augments)




import curses
from .element import Element

class RowBar(Element):
    """An element container for displaying multiple elements inline in a menu.
    :param elements: A list of Elements that will be added to the RowBar
    :type elements: list
    :param separator: The character(s) that will be printed between each element in the bar
    :type separator: str, optional
    """
    def __init__(self, elements, separator='\t', color=0,):
        """Constructor method
        """
        super().__init__('', color=color)
        self.selectable = True 
        self.isContainer = True
        self.elements = elements
        self.separator = separator
        self.selectedIndex = 0

        allNotSelectable = True
        for element in self.elements:
            if element.selectable:
                allNotSelectable = False
                break

        if allNotSelectable:
            self.selectable = False

        # Shortcut to ensure that an element is selected by default to prevent selecting
        # unselectable elements
        if self.selectable:
            self.increaseSelectedIndex()
            self.decreaseSelectedIndex()


    def handleInput(self, key):
        """Pass input keys from the menu to the selected element
        :param key: The key that was input
        :type key: int
        """
        if (key in (curses.KEY_LEFT, ord('h'))):
            self.decreaseSelectedIndex()
        elif (key in (curses.KEY_RIGHT, ord('l'))):
            self.increaseSelectedIndex()

    def getStr(self, selected=False):
        """Get the display string of the rowbar, including getting the strings of all of 
        the elements contained inside the rowbar
        :param selected: Whether or not the bar is selected in the menu
        :type selected: bool
        :return: The text that the bar should be printed as
        :rtype: str
        """
        string = ''
        for id, element in enumerate(self.elements):
            if id == self.selectedIndex and selected:
                string += element.getStr(selected=True) + self.separator
            else:
                string += element.getStr() + self.separator

        return string

    def triggerAction(self):
        """Trigger the action of the element contained in the bar at the selected index"""
        self.elements[self.selectedIndex].triggerAction()

    def increaseSelectedIndex(self):
        """Increase the selected index of the menu, skipping unselectable elements"""
        if self.selectedIndex >= len(self.elements)-1:
            return

        self.selectedIndex += 1
        while self.elements[self.selectedIndex].selectable == False:
            self.increaseSelectedIndex()
            if(self.selectedIndex >= len(self.elements)-1):
                self.decreaseSelectedIndex()
                break

    def decreaseSelectedIndex(self):
        """Decrease the selected index of the menu, skipping unselectable elements"""
        if self.selectedIndex <= 0:
            return
        self.selectedIndex -= 1
        while self.elements[self.selectedIndex].selectable == False:
            self.decreaseSelectedIndex()
            if(self.selectedIndex <= 0):
                self.increaseSelectedIndex()
                break

    def update(self):
        """Update each element in the RowBar"""
        for element in self.elements:
            element.update()




from .element import Element

class CheckBox(Element):
    """This class is a UI element that works as a checkbox. It can be toggled, 
    and its current state can be easily read as either True of False"""

    def __init__(self, label, action=None, refreshFuncton=None, color=0, state=False):
        self.label = label
        self.action = action
        self.refreshFunction = refreshFuncton
        self.selectable = True
        self.color = color
        self.state = state

    




class Element:
    """A parent class to base all UI elements off of. A base element draws similar
    to a label, but it is not recommended to insert a plain element into a menu
    :param label: The text to display on the element
    :type label: str
    :param refreshFunction: The function to run to refresh the label of the element
    :type refreshFunction: function
    :param color: The color to draw the text of the element in
    :type color: int, optional
    """
    def __init__(self, label, refreshFunction=None, color=0):
        """Constructor method
        """
        self.label = label
        self.color = color
        self.refreshFunction = refreshFunction
        self.selectable = False
        self.isContainer = False
    
    def update(self):
        """Update the label of the element by running its refreshFunction
        """
        if self.refreshFunction == None:
            return 

        try:
            self.label = self.refreshFunction()
        except:
            raise Exception("Unable to run refreshFunction for label:", self.label)
    
    def triggerAction(self):
        raise NotImplementedError("This method should be overwritten by subclasses")

    def getStr(self, selected=False):
        """Get the display string of the element
        :return: The text to display the element as
        :rtype: str
        """
        return self.label
    
    def handleInput(self, key):
        pass




from .uiManager import UIManager
from .element import Element
from .menu import Menu
from .button import Button
from .label import Label
from .rowBar import RowBar
from .progressBar import ProgressBar



import curses
import time

class UIManager:
    """A class for managing and running all UI functions. The UIManager stores a dict of
    menus to easily switch between them, and passes keyboard inputs to the selected menu only,
    as well as updating the menu
    :param stdscr: The screen to print menus to
    """
    def __init__(self, stdscr):
        self.stdscr = stdscr
        self.currentMenu = None
        self.menus = {} 
        self.shouldExit = False

        curses.curs_set(0)
        curses.start_color()
        curses.use_default_colors()

        # Background color pallette (XTerm-256)
        bgColors = [-1,0,1,2,3,4,5,6,236]

        for i in range(len(bgColors)):
            for n in range(16):
                curses.init_pair(i*16+n, n-1, bgColors[i])


        # Initialize color pairs
        #for i in range(1, min(curses.COLORS-2, 128+16)):
        #    curses.init_pair(i,(i%16)-1, bgColors[(i//16)-1])
        #try:
        #    curses.init_pair(0, -1, -1)
        #except:
        #    # TODO: Fix windows color schemes
        #    pass

        self.stdscr.bkgd(' ', curses.color_pair(0))
        self.stdscr.clear()
        self.stdscr.refresh()
        self.stdscr.timeout(100)

    def run(self):
        """Reset the current display and run the main loop"""
        self.stdscr.clear()
        self.stdscr.refresh()
        self.mainLoop()

    def mainLoop(self):
        """Continually update menus and elements until the program exits"""
        try:
            while self.shouldExit == False:
                if self.currentMenu != None:
                    self.stdscr.clear()
                    self.menus[self.currentMenu].update()
                    self.menus[self.currentMenu].display(self.stdscr)
                key = self.stdscr.getch()
                self.menus[self.currentMenu].handleInput(key)
        except KeyboardInterrupt:
            self.shouldExit = False 

    def addMenu(self, menu):
        """Insert a menu into the dictionary of menus. Note that menus can
        be overwritten without warning.
        :param menu: The menu to insert
        :type menu: ui.Menu
        """
        name = menu.name

        self.menus[name] = menu

    def switchMenu(self, menu):
        """Switch the currently displayed menu
        :param menu: The name of the menu to switch to
        :type menu: str
        """
        if menu in self.menus.keys():
            self.currentMenu = menu

